package ch.zhaw.pm2.racetrack.strategy;

import ch.zhaw.pm2.racetrack.PositionVector;
import ch.zhaw.pm2.racetrack.Track;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

public class PathFinderMoveStrategy implements MoveStrategy{
    private Track track;
    private int carIndex;
    private List<PositionVector.Direction> moveList;
    private int pointer;
    private List<PositionVector.Direction> allDirections;
    private List<State> calculatedStates;


    public PathFinderMoveStrategy(Track track, int carIndex) {
        this.track = track;
        this.carIndex = carIndex;
        allDirections = Arrays.asList(PositionVector.Direction.values());
        createMoveList();
        pointer = -1;
    }


    private void createMoveList(){
        calculatedStates = new ArrayList<>();
        PossibleMove finishedMove = null;
        List<PossibleMove> possibleMoves= new ArrayList<>();
        for(PositionVector.Direction direction : allDirections){
            PossibleMove newMove = new PossibleMove(null, direction);
            if(! newMove.crashed()){
                possibleMoves.add(newMove);
            }

        }
        while(finishedMove == null){
            List<PossibleMove> newMoves = new ArrayList<>();
            for(PossibleMove previousMove : possibleMoves){
                for(PositionVector.Direction direction : allDirections){
                    PossibleMove newMove = new PossibleMove(previousMove, direction);
                    State newState = new State(newMove.endPosition, newMove.endVelocity);
                        if(! (newMove.crashed() || alreadyCalculated(newState) || finishedMove != null)){
                            if(newMove.finished()){
                                finishedMove = newMove;
                            } else {
                                calculatedStates.add(newState);
                                newMoves.add(newMove);
                            }
                        }
                }
            }
            possibleMoves = newMoves;

        }


        moveList = finishedMove.directions;


    }

    private boolean alreadyCalculated(State state){
        for(State calculatedState: calculatedStates){
            if(state.equals(calculatedState)){
                return true;
            }
        }
        return false;
    }

    public class State{
        PositionVector position;
        PositionVector velocity;

        public State(PositionVector position, PositionVector velocity){
            this.position = position;
            this.velocity = velocity;
        }

        public boolean equals(State compareState){
            if(compareState.position.equals(position) && compareState.velocity.equals(velocity)){
                return true;
            } else{
                return false;
            }
        }
    }


    public class PossibleMove {
        List<PositionVector.Direction> directions;
        PositionVector startPosition;
        PositionVector endPosition;
        PositionVector endVelocity;


        public PossibleMove(PossibleMove previousMove, PositionVector.Direction nextDirection){
            PositionVector startVelocity;

            directions = new ArrayList<>();
            if(previousMove != null){
                directions.addAll(previousMove.directions);
                startPosition = previousMove.endPosition;
                startVelocity = previousMove.endVelocity;
            }
            else {
                startPosition = track.getCarPos(carIndex);
                startVelocity = track.getCar(carIndex).getVelocity();
            }
            directions.add(nextDirection);
            endVelocity = new PositionVector(startVelocity.getX() + nextDirection.vector.getX(), startVelocity.getY() + nextDirection.vector.getY());
            endPosition = new PositionVector(startPosition.getX() + endVelocity.getX(), startPosition.getY() + endVelocity.getY());
        }

        public boolean finished(){
            if(track.calculateNewWinPoints(startPosition, endPosition) == 1){
                return true;
            }
            else{
                return false;
            }
        }

        public boolean crashed() {
            List<PositionVector> points = track.calculatePointsOnPath(startPosition, endPosition);
            for(PositionVector point : points) {
                if (track.willCrashAtPosition(carIndex, point)){
                    return true;
                }
            }
            if(track.calculateNewWinPoints(startPosition, endPosition) == -1){
                return true;
            }
            return false;
        }

    }



    @Override
    public PositionVector.Direction nextMove() { //TODO check for crash and recreate movelist if crash
        pointer += 1;
        if (pointer < moveList.size()) {
            return moveList.get(pointer);
        }
        return null;
    }


}
